from rest_framework import exceptions


class PermissionRequiredMixin:

    perms_map = {
        'GET': [],
        'OPTIONS': [],
        'HEAD': [],
        # VIEW method is assigned when actual request method
        # is GET and a database record object was found
        'LIST': ['%(app_label)s.list_%(model_name)s'],
        'VIEW': ['%(app_label)s.view_%(model_name)s'],
        'POST': ['%(app_label)s.add_%(model_name)s'],
        'PUT': ['%(app_label)s.change_%(model_name)s'],
        'PATCH': ['%(app_label)s.change_%(model_name)s'],
        'DELETE': ['%(app_label)s.delete_%(model_name)s'],
    }

    object_permission_required = None
    permission_required = None

    def confirm_authentication(self, request):
        """
        Check if user is authenticated before doing other permission checks.
        Can be overridden on subclasses to refine logic for auth requirements.
        """
        if not request.user.is_authenticated:
            self.permission_denied(request)

    def get_required_permissions(self, method, model_cls):
        """
        Given a model and an HTTP method, return the list of permission
        codes that the user is required to have.
        """
        kwargs = {
            'app_label': model_cls._meta.app_label,
            'model_name': model_cls._meta.model_name,
        }

        if method not in self.perms_map:
            raise exceptions.MethodNotAllowed(method)

        return [perm % kwargs for perm in self.perms_map[method]]

    def check_object_permissions(self, request, obj):

        self.confirm_authentication(request)
        user = request.user

        # Set perms for Detail pages, but not list pages, where no object is present
        # won't always be able to refine query to prevent lookups
        method = 'VIEW' if request.method == 'GET' else request.method
        required_perms = self.get_required_permissions(method, obj)

        missing_permissions = [perm for perm in required_perms if not user.has_perm(perm, obj)]
        if any(missing_permissions):

            # Return 404 so the user can't determine if the object even exists
            if method == 'VIEW' or method == 'HEAD':
                raise exceptions.NotFound()

            self.permission_denied(
                request, message=f'Missing Perms: {", ".join(missing_permissions)}'
            )

    def check_permissions(self, request):

        self.confirm_authentication(request)
        user = request.user
        model_obj = self.queryset.model

        try:
            is_list_view = request.resolver_match.url_name.endswith(('-list', '_list'))
        except AttributeError:
            # url_name may not be set in some test-framework situations
            is_list_view = False

        method = 'LIST' if is_list_view and request.method in ('GET', 'HEAD') else request.method
        required_perms = self.get_required_permissions(method, model_obj)

        missing_permissions = [perm for perm in required_perms if not user.has_perm(perm)]
        if any(missing_permissions):
            self.permission_denied(
                request, message=f'Missing Perms: {", ".join(missing_permissions)}'
            )


__all__ = [
    'PermissionRequiredMixin',
]
